from datetime import datetime

from cpl_core.configuration import ConfigurationABC
from cpl_core.logging import LoggerABC
from cpl_discord.events import OnReadyABC
from cpl_discord.service import DiscordBotServiceABC
from cpl_translation import TranslatePipe
from discord import guild

from bot_core.abc.message_service_abc import MessageServiceABC
from bot_core.configuration.feature_flags_enum import FeatureFlagsEnum
from bot_core.configuration.feature_flags_settings import FeatureFlagsSettings
from bot_data.model.server_config import ServerConfig
from bot_data.model.technician_config import TechnicianConfig


class BootLogOnReadyEvent(OnReadyABC):
    def __init__(
        self,
        config: ConfigurationABC,
        logger: LoggerABC,
        bot: DiscordBotServiceABC,
        message_service: MessageServiceABC,
        translate: TranslatePipe,
        tech_config: TechnicianConfig,
    ):
        OnReadyABC.__init__(self)
        self._config = config

        self._logger = logger
        self._bot = bot
        self._message_service = message_service
        self._t = translate
        self._tech_config = tech_config

        self._logger.info(__name__, f"Module {type(self)} loaded")

    async def on_ready(self):
        self._logger.debug(__name__, f"Module {type(self)} started")
        if not FeatureFlagsSettings.get_flag_from_dict(
            self._tech_config.feature_flags, FeatureFlagsEnum.boot_log_module
        ):
            return

        try:
            start_time = self._config.get_configuration("Bot_StartTime")
            init_time = round(
                (datetime.now() - datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S.%f")).total_seconds(),
                2,
            )
            self._config.add_configuration("InitTime", str(init_time))
            self._logger.debug(__name__, f"Bot Init time:  {init_time}s")
            # print warning if initialisation took too long
            if init_time >= 30:
                self._logger.warn(__name__, "It takes too long to start the bot!")

            # print error if initialisation took way too long
            elif init_time >= 90:
                self._logger.error(__name__, "It takes far too long to start the bot!!!")
        except Exception as e:
            self._logger.error(__name__, "Init time calculation failed", e)
            return

        for g in self._bot.guilds:
            g: guild = g
            self._logger.debug(__name__, f"Server detected: {g.id}")

            server_config: ServerConfig = self._config.get_configuration(f"ServerConfig_{g.id}")
            if server_config is None:
                self._logger.error(__name__, f"Config ServerConfig_{g.id} not found!")
                return

            if not FeatureFlagsSettings.get_flag_from_dict(
                server_config.feature_flags, FeatureFlagsEnum.boot_log_module
            ):
                continue

            self._bot.loop.create_task(
                self._message_service.send_channel_message(
                    self._bot.get_channel(server_config.login_message_channel_id),
                    self._t.transform("modules.boot_log.login_message").format(init_time),
                )
            )

        self._config.add_configuration("IS_READY", "true")
        self._logger.info(__name__, "Bot is ready")
        self._logger.trace(__name__, f"Module {type(self)} stopped")
