import { Component, OnInit } from '@angular/core';
import { FormBuilder, FormGroup, Validators } from '@angular/forms';
import { Router } from '@angular/router';
import { TranslateService } from '@ngx-translate/core';
import { catchError } from 'rxjs/operators';
import { AuthErrorMessages } from 'src/app/models/auth/auth-error-messages.enum';
import { UpdateUserDTO } from 'src/app/models/auth/update-user.dto';
import { ErrorDTO } from 'src/app/models/error/error-dto';
import { ServiceErrorCode } from 'src/app/models/error/service-error-code.enum';
import { AuthService } from 'src/app/services/auth/auth.service';
import { SpinnerService } from 'src/app/services/spinner/spinner.service';
import { ToastService } from 'src/app/services/toast/toast.service';
import { throwError } from "rxjs";

@Component({
  selector: 'app-change-password',
  templateUrl: './change-password.component.html',
  styleUrls: ['./change-password.component.scss']
})
export class ChangePasswordComponent implements OnInit {

  passwordForm!: FormGroup;
  submitted = false;
  ready = false;
  errors = {
    email: false,
    repeatPassword: false,
    oldPassword: false
  };

  constructor(
    private authService: AuthService,
    private formBuilder: FormBuilder,
    private router: Router,
    private spinnerService: SpinnerService,
    private toastService: ToastService,
    private translate: TranslateService
  ) { }

  ngOnInit(): void {
    this.initForms();
  }

  initForms(): void {
    this.passwordForm = this.formBuilder.group({
      oldPassword: [null, [Validators.required, Validators.minLength(8)]],
      password: [null, [Validators.required, Validators.minLength(8)]],
      passwordRepeat: [null, [Validators.required, Validators.minLength(8)]]
    });
  }

  resetErrorFlags(): void {
    this.errors = {
      email: false,
      repeatPassword: false,
      oldPassword: false
    };
  }

  changePassword(): void {
    this.submitted = true;
    this.resetErrorFlags();

    if (this.passwordForm.value.password !== this.passwordForm.value.passwordRepeat) {
      this.errors.repeatPassword = true;
      return;
    }
    this.spinnerService.showSpinner();

    const decodedToken = this.authService.getDecodedToken();
    const changePasswordDTO: UpdateUserDTO = {
      authUserDTO: {
        firstName: null,
        lastName: null,
        password: this.passwordForm.value.oldPassword,
        email: this.authService.getEMailFromDecodedToken(decodedToken)
      },
      newAuthUserDTO: {
        firstName: null,
        lastName: null,
        password: this.passwordForm.value.password,
        email: this.authService.getEMailFromDecodedToken(decodedToken)
      }
    };

    this.authService.updateUser(changePasswordDTO)
      .pipe(catchError(error => {
        if (error.error !== null) {
          const err: ErrorDTO = error.error;

          if (err.errorCode === ServiceErrorCode.InvalidUser && err.message === AuthErrorMessages.WrongPassword) {
            this.errors.oldPassword = true;
          }

          error.error = null;
        }
        this.spinnerService.hideSpinner();
        this.toastService.error(this.translate.instant('view.change_password.message.error'), this.translate.instant('view.change_password.message.password_cannot_be_changed'));

        return throwError(() => error);
      }))
      .subscribe(resp => {
        this.spinnerService.hideSpinner();
        this.toastService.success(this.translate.instant('view.change_password.message.change_password'), this.translate.instant('view.change_password.message.changed_password'));
        this.router.navigate(["/dashboard"]);
      });
  }

}
