import { Component, OnDestroy, OnInit } from "@angular/core";
import { AuthService } from "../../../../../../services/auth/auth.service";
import { SpinnerService } from "../../../../../../services/spinner/spinner.service";
import { ToastService } from "../../../../../../services/toast/toast.service";
import { ConfirmationDialogService } from "../../../../../../services/confirmation-dialog/confirmation-dialog.service";
import { FormBuilder, FormControl, FormGroup } from "@angular/forms";
import { TranslateService } from "@ngx-translate/core";
import { DataService } from "../../../../../../services/data/data.service";
import { SidebarService } from "../../../../../../services/sidebar/sidebar.service";
import { ActivatedRoute } from "@angular/router";
import { Page } from "../../../../../../models/graphql/filter/page.model";
import { Sort, SortDirection } from "../../../../../../models/graphql/filter/sort.model";
import { Level, LevelFilter } from "../../../../../../models/data/level.model";
import { LevelListQuery, Query } from "../../../../../../models/graphql/query.model";
import { Queries } from "../../../../../../models/graphql/queries.model";
import { catchError, debounceTime, takeUntil } from "rxjs/operators";
import { LazyLoadEvent } from "primeng/api";
import { Table } from "primeng/table";
import { User } from "../../../../../../models/data/user.model";
import { LevelMutationResult, UpdateUserMutationResult } from "../../../../../../models/graphql/result.model";
import { Mutations } from "../../../../../../models/graphql/mutations.model";
import { forkJoin, Subject, throwError } from "rxjs";
import { Server } from "../../../../../../models/data/server.model";
import { UserDTO } from "../../../../../../models/auth/auth-user.dto";
import { ComponentWithTable } from "../../../../../../base/component-with-table";

@Component({
  selector: "app-levels",
  templateUrl: "./levels.component.html",
  styleUrls: ["./levels.component.scss"]
})
export class LevelsComponent extends ComponentWithTable implements OnInit, OnDestroy {
  public levels: Level[] = [];
  public loading = true;

  public filterForm!: FormGroup<{
    id: FormControl<number | null>,
    name: FormControl<string | null>,
    color: FormControl<string | null>,
    min_xp: FormControl<number | null>,
    permissions: FormControl<string | null>,
  }>;

  public filter: LevelFilter = {};
  public page: Page = {
    pageSize: undefined,
    pageIndex: undefined
  };
  public sort: Sort = {
    sortColumn: undefined,
    sortDirection: undefined
  };

  public totalRecords: number = 0;

  public clonedLevels: { [s: string]: Level; } = {};

  private unsubscriber = new Subject<void>();
  private server: Server = {};
  public user: UserDTO | null = null;

  query: string = Queries.levelWithHistoryQuery;

  public constructor(
    private authService: AuthService,
    private spinner: SpinnerService,
    private toastService: ToastService,
    private confirmDialog: ConfirmationDialogService,
    private fb: FormBuilder,
    private translate: TranslateService,
    private data: DataService,
    private sidebar: SidebarService,
    private route: ActivatedRoute
  ) {
    super("level", ["id", "name", "color", "min_xp", "permissions"], (oldElement: Level, newElement: Level) => {
      return oldElement.name === newElement.name;
    });
  }

  public ngOnInit(): void {
    this.setFilterForm();
    this.data.getServerFromRoute(this.route).then(async server => {
      this.server = server;
      this.loadNextPage();
      let authUser = await this.authService.getLoggedInUser();
      this.user = authUser?.users?.find(u => u.server == this.server.id) ?? null;
    });
  }

  public ngOnDestroy(): void {
    this.unsubscriber.next();
    this.unsubscriber.complete();
  }

  public loadNextPage(): void {
    this.loading = true;
    this.data.query<LevelListQuery>(Queries.levelQuery, {
        serverId: this.server.id, filter: this.filter, page: this.page, sort: this.sort
      },
      (data: Query) => {
        return data.servers[0];
      }
    ).subscribe(data => {
      this.totalRecords = data.levelCount;
      this.levels = data.levels;
      this.spinner.hideSpinner();
      this.loading = false;
    });
  }

  public setFilterForm(): void {
    this.filterForm = this.fb.group({
      id: new FormControl<number | null>(null),
      name: new FormControl<string | null>(null),
      color: new FormControl<string | null>(null),
      min_xp: new FormControl<number | null>(null),
      permissions: new FormControl<string | null>(null)
    });

    this.filterForm.valueChanges.pipe(
      takeUntil(this.unsubscriber),
      debounceTime(600)
    ).subscribe(changes => {
      if (changes.id) {
        this.filter.id = changes.id;
      } else {
        this.filter.id = undefined;
      }

      if (changes.name) {
        this.filter.name = changes.name;
      } else {
        this.filter.name = undefined;
      }

      if (this.page.pageSize)
        this.page.pageSize = 10;

      if (this.page.pageIndex)
        this.page.pageIndex = 0;

      this.loadNextPage();
    });
  }

  public newLevelTemplate: Level = {
    createdAt: "",
    modifiedAt: ""
  };

  public nextPage(event: LazyLoadEvent): void {
    this.page.pageSize = event.rows ?? 0;
    if (event.first != null && event.rows != null)
      this.page.pageIndex = event.first / event.rows;
    this.sort.sortColumn = event.sortField ?? undefined;
    this.sort.sortDirection = event.sortOrder === 1 ? SortDirection.ASC : event.sortOrder === -1 ? SortDirection.DESC : SortDirection.ASC;

    this.loadNextPage();
  }

  public resetFilters(): void {
    this.filterForm.reset();
  }

  public onRowEditInit(table: Table, user: User, index: number): void {
    this.clonedLevels[index] = { ...user };
  }

  public override onRowEditSave(newLevel: Level, index: number): void {
    if (this.isEditingNew && JSON.stringify(newLevel) === JSON.stringify(this.newLevelTemplate)) {
      this.isEditingNew = false;
      this.levels.splice(index, 1);
      return;
    }

    if (!newLevel.id && !this.isEditingNew || !newLevel.minXp && !newLevel?.name && !newLevel?.permissions) {
      return;
    }

    if (this.isEditingNew) {
      this.spinner.showSpinner();
      this.data.mutation<LevelMutationResult>(Mutations.createLevel, {
          name: newLevel.name,
          color: newLevel.color,
          minXp: newLevel.minXp,
          permissions: newLevel.permissions,
          serverId: this.server.id
        }
      ).pipe(catchError(err => {
        this.isEditingNew = false;
        this.spinner.hideSpinner();
        return throwError(err);
      })).subscribe(result => {
        this.isEditingNew = false;
        this.spinner.hideSpinner();
        this.toastService.success(this.translate.instant("view.server.levels.message.level_create"), this.translate.instant("view.server.levels.message.level_create_d", { name: result.level.createLevel?.name }));
        this.loadNextPage();
      });
      return;
    }

    this.spinner.showSpinner();
    this.data.mutation<UpdateUserMutationResult>(Mutations.updateLevel, {
        id: newLevel.id,
        name: newLevel.name,
        color: newLevel.color,
        minXp: newLevel.minXp,
        permissions: newLevel.permissions
      }
    ).pipe(catchError(err => {
      this.spinner.hideSpinner();
      return throwError(err);
    })).subscribe(_ => {
      this.spinner.hideSpinner();
      this.toastService.success(this.translate.instant("view.server.levels.message.level_update"), this.translate.instant("view.server.levels.message.level_update_d", { name: newLevel.name }));
      this.loadNextPage();
    });

  }

  public onRowEditCancel(index: number): void {
    if (this.isEditingNew) {
      this.levels.splice(index, 1);
      delete this.clonedLevels[index];
      this.isEditingNew = false;
      return;
    }

    this.levels[index] = this.clonedLevels[index];
    delete this.clonedLevels[index];
  }

  public deleteLevel(level: Level): void {
    this.confirmDialog.confirmDialog(
      this.translate.instant("view.server.levels.message.level_delete"), this.translate.instant("view.server.levels.message.level_delete_q", { name: level.name }),
      () => {
        this.spinner.showSpinner();
        this.data.mutation<LevelMutationResult>(Mutations.deleteLevel, {
            id: level.id
          }
        ).pipe(catchError(err => {
          this.spinner.hideSpinner();
          return throwError(err);
        })).subscribe(l => {
          this.spinner.hideSpinner();
          this.toastService.success(this.translate.instant("view.server.levels.message.level_deleted"), this.translate.instant("view.server.levels.message.level_deleted_d", { name: level.name }));
          this.loadNextPage();
        });
      });
  }

  public addLevel(table: Table): void {
    const newLevel = JSON.parse(JSON.stringify(this.newLevelTemplate));

    this.levels = [newLevel, ...this.levels];

    table.initRowEdit(newLevel);

    const index = this.levels.findIndex(l => l.id == newLevel.id);
    this.onRowEditInit(table, newLevel, index);

    this.isEditingNew = true;
  }

}
