import { Component, OnDestroy, OnInit } from "@angular/core";
import { ActivatedRoute, Router } from "@angular/router";
import { Queries } from "../../../../models/graphql/queries.model";
import { LevelListQuery, Query, UserListQuery, UserWarningQuery } from "../../../../models/graphql/query.model";
import { SpinnerService } from "../../../../services/spinner/spinner.service";
import { DataService } from "../../../../services/data/data.service";
import { User } from "../../../../models/data/user.model";
import { UserDTO } from "../../../../models/auth/auth-user.dto";
import { AuthService } from "src/app/services/auth/auth.service";
import { ToastService } from "src/app/services/toast/toast.service";
import { TranslateService } from "@ngx-translate/core";
import { Server } from "../../../../models/data/server.model";
import { forkJoin, Subject, throwError } from "rxjs";
import { catchError, takeUntil } from "rxjs/operators";
import { Table } from "primeng/table";
import { UpdateUserMutationResult } from "../../../../models/graphql/result.model";
import { Mutations } from "../../../../models/graphql/mutations.model";
import { MenuItem } from "primeng/api";
import { UserWarning } from "../../../../models/data/user_warning.model";
import moment from "moment";

@Component({
  selector: "app-profile",
  templateUrl: "./profile.component.html",
  styleUrls: ["./profile.component.scss"]
})
export class ProfileComponent implements OnInit, OnDestroy {

  user: User = { createdAt: "", modifiedAt: "" };
  levels!: MenuItem[];
  private server: Server = {};
  private author?: UserDTO;
  private clonedUserWarnings: UserWarning[] = [];
  public isEditingNewUserWarning: boolean = false;
  public isEditing: boolean = false;
  public isModerator: boolean = false;

  private unsubscriber = new Subject<void>();

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private spinner: SpinnerService,
    private data: DataService,
    private auth: AuthService,
    private toast: ToastService,
    private translate: TranslateService,
    private toastService: ToastService
  ) {
  }

  public ngOnInit(): void {
    this.isEditing = false;
    this.loadProfile();
  }

  private loadProfile() {
    this.route.params.pipe(takeUntil(this.unsubscriber)).subscribe(params => {
      this.data.getServerFromRoute(this.route).then(async (server) => {
        if (!params["memberId"] || params["memberId"] == "undefined") {
          await this.router.navigate([`/server/${server.id}`]);
          return;
        }
        this.server = server;

        this.data.query<LevelListQuery>(Queries.levelQuery, {
            serverId: server.id
          },
          (data: Query) => {
            return data.servers[0];
          }
        ).subscribe(data => {
          this.levels = data.levels.map(level => {
            return { label: level.name, value: level };
          });
        });

        let authUser = await this.auth.getLoggedInUser();
        this.spinner.showSpinner();
        let user: UserDTO | null = authUser?.users?.find(u => u.server == server.id) ?? null;
        if (!user || user?.id != params["memberId"] && !user?.isModerator) {
          this.toast.error(this.translate.instant("view.server.profile.permission_denied"), this.translate.instant("view.server.profile.permission_denied_d"));
          this.spinner.hideSpinner();
          await this.router.navigate(["/server", server.id]);
          return;
        }
        this.author = user;
        this.isModerator = user?.isModerator;

        this.data.query<UserListQuery>(Queries.userProfile, {
            serverId: this.server.id,
            userId: params["memberId"]
          }
        ).subscribe(users => {
          if (!users.users[0]) {
            this.router.navigate([`/server/${server.id}`]);
          }
          this.user = users.users[0];

          this.data.query<UserWarningQuery>(Queries.userProfileWarnings, {
              serverId: this.server.id,
              userId: this.user.id
            },
            (data: UserListQuery) => {
              return data.users[0];
            }
          ).subscribe(result => {
            this.user.userWarningCount = result.userWarningCount;
            this.user.userWarnings = result.userWarnings;

            this.spinner.hideSpinner();
          });
        });
      });
    });
  }

  public toogleEditUser() {
    this.isEditing = !this.isEditing;
  }

  public updateUser() {
    this.spinner.showSpinner();
    this.data.mutation<UpdateUserMutationResult>(Mutations.updateUser, {
        id: this.user.id,
        xp: this.user.xp,
        birthday: moment(this.user.birthday).format("DD.MM.YYYY"),
        levelId: this.user.level?.id,
        userWarnings: this.user.userWarnings?.map(userWarning => {
          return {
            id: userWarning.id,
            user: userWarning.user?.id ?? this.user.id,
            description: userWarning.description,
            author: userWarning.author?.id ?? this.author?.id
          };
        })
      }
    ).pipe(catchError(err => {
      this.spinner.hideSpinner();
      this.isEditing = false;
      return throwError(err);
    })).subscribe(_ => {
      this.spinner.hideSpinner();
      this.toastService.success(this.translate.instant("view.server.members.message.user_changed"), this.translate.instant("view.server.members.message.user_changed_d", { name: this.user.name }));
      this.isEditing = false;
      this.loadProfile();
    });
  }

  public ngOnDestroy(): void {
    this.unsubscriber.next();
    this.unsubscriber.complete();
  }

  public onBeforeToggle(event: Event, collapsed: boolean) {
    const filterUser = (x: { users: User[] }) => {
      const users = x.users ?? [];
      return users[0];
    };

    if (collapsed) {
      this.spinner.showSpinner();
      forkJoin([
        this.data.query<User>(Queries.userProfileAchievements, {
            serverId: this.server.id,
            userId: this.user.id
          },
          filterUser
        ),
        this.data.query<User>(Queries.userProfileVoiceChannelJoins, {
            serverId: this.server.id,
            userId: this.user.id
          },
          filterUser
        ),
        this.data.query<User>(Queries.userProfileGameserverJoins, {
            serverId: this.server.id,
            userId: this.user.id
          },
          filterUser
        )
      ]).subscribe(data => {
        this.user.achievementCount = data[0].achievementCount;
        this.user.achievements = data[0].achievements;

        this.user.joinedVoiceChannelCount = data[1].joinedVoiceChannelCount;
        this.user.joinedVoiceChannels = data[1].joinedVoiceChannels;

        this.user.userJoinedGameServerCount = data[2].userJoinedGameServerCount;
        this.user.userJoinedGameServers = data[2].userJoinedGameServers;
        this.spinner.hideSpinner();
      });
      return;
    }
    this.user.achievementCount = 0;
    this.user.achievements = [];

    this.user.userJoinedGameServerCount = 0;
    this.user.userJoinedGameServers = [];

    this.user.joinedVoiceChannelCount = 0;
    this.user.joinedVoiceChannels = [];
  }

  public addNewUserWarning(table: Table) {
    const newWarning: UserWarning = {
      description: "",
      user: this.user
    };

    this.user.userWarnings = [newWarning, ...this.user.userWarnings ?? []];

    table.initRowEdit(newWarning);

    const index = this.user.userWarnings.findIndex(l => l.id == newWarning.id);
    this.onRowEditInit(table, newWarning, index);

    this.isEditingNewUserWarning = true;
  }

  public onRowEditInit(table: Table, user: User, index: number): void {
    this.clonedUserWarnings[index] = { ...user };
  }

  public deleteUserWarning(index: number) {
    this.user.userWarnings?.splice(index, 1);
  }

  public editSaveUserWarning(value: any, index: number) {
    this.isEditingNewUserWarning = false;
    if (!value.value || !this.user.userWarnings || this.user.userWarnings[index] == this.clonedUserWarnings[index]) {
      return;
    }

    delete this.clonedUserWarnings[index];
  }

  public editCancelUserWarning(index: number) {
    if (this.user.userWarnings) {
      this.user.userWarnings[index] = this.clonedUserWarnings[index];
    }
    delete this.clonedUserWarnings[index];
  }
}
