import { Component, OnDestroy, OnInit } from "@angular/core";
import { AuthService } from "../../../../../../services/auth/auth.service";
import { SpinnerService } from "../../../../../../services/spinner/spinner.service";
import { ToastService } from "../../../../../../services/toast/toast.service";
import { ConfirmationDialogService } from "../../../../../../services/confirmation-dialog/confirmation-dialog.service";
import { FormBuilder, FormControl, FormGroup } from "@angular/forms";
import { TranslateService } from "@ngx-translate/core";
import { DataService } from "../../../../../../services/data/data.service";
import { SidebarService } from "../../../../../../services/sidebar/sidebar.service";
import { ActivatedRoute } from "@angular/router";
import { Server } from "../../../../../../models/data/server.model";
import { ComponentWithTable } from "../../../../../../base/component-with-table";
import { ShortRoleName, ShortRoleNameFilter } from "../../../../../../models/data/short_role_name.model";
import { Page } from "../../../../../../models/graphql/filter/page.model";
import { Sort, SortDirection } from "../../../../../../models/graphql/filter/sort.model";
import { Subject, throwError } from "rxjs";
import { UserDTO } from "../../../../../../models/auth/auth-user.dto";
import { Queries } from "../../../../../../models/graphql/queries.model";
import { ShortRoleNameListQuery, Query, SingleDiscordQuery, ShortRoleNamePositionsQuery } from "../../../../../../models/graphql/query.model";
import { catchError, debounceTime, takeUntil } from "rxjs/operators";
import { LazyLoadEvent, MenuItem } from "primeng/api";
import { Table } from "primeng/table";
import { User } from "../../../../../../models/data/user.model";
import { ShortRoleNameMutationResult, UpdateUserMutationResult } from "../../../../../../models/graphql/result.model";
import { Mutations } from "../../../../../../models/graphql/mutations.model";

@Component({
  selector: "app-short-role-names",
  templateUrl: "./short-role-names.component.html",
  styleUrls: ["./short-role-names.component.scss"]
})
export class ShortRoleNamesComponent extends ComponentWithTable implements OnInit, OnDestroy {
  public shortRoleNames: ShortRoleName[] = [];
  public loading = true;

  public filterForm!: FormGroup<{
    id: FormControl<number | null>,
    shortName: FormControl<string | null>,
    roleId: FormControl<string | null>,
    position: FormControl<string | null>,
  }>;

  public filter: ShortRoleNameFilter = {};
  public page: Page = {
    pageSize: undefined,
    pageIndex: undefined
  };
  public sort: Sort = {
    sortColumn: undefined,
    sortDirection: undefined
  };

  public totalRecords: number = 0;

  public clonedShortRoleNames: { [s: string]: ShortRoleName; } = {};

  private unsubscriber = new Subject<void>();
  private server: Server = {};
  public user: UserDTO | null = null;

  query: string = Queries.shortRoleNameWithHistoryQuery;
  roles: MenuItem[] = [];
  positions: MenuItem[] = [];

  public constructor(
    private authService: AuthService,
    private spinner: SpinnerService,
    private toastService: ToastService,
    private confirmDialog: ConfirmationDialogService,
    private fb: FormBuilder,
    private translate: TranslateService,
    private data: DataService,
    private sidebar: SidebarService,
    private route: ActivatedRoute
  ) {
    super("short-role-names", ["id", "name", "role", "position"], (oldElement: ShortRoleName, newElement: ShortRoleName) => {
      return oldElement.shortName === newElement.shortName &&
        oldElement.roleId === newElement.roleId;
    });
  }

  public ngOnInit(): void {
    this.setFilterForm();
    this.data.getServerFromRoute(this.route).then(async server => {
      this.server = server;
      this.loadNextPage();

      this.spinner.showSpinner();
      this.data.query<SingleDiscordQuery>(Queries.guildsQuery, {
          id: server?.discordId
        }
      ).subscribe(data => {
        if (data.discord.guilds) {
          this.roles = data.discord.guilds[0].roles
            .map(x => {
              return { label: x.name, value: x };
            });
        }
        this.spinner.hideSpinner();
      });
      this.data.query<ShortRoleNamePositionsQuery>(Queries.shortRoleNamePositionsQuery).subscribe(data => {
        if (data.shortRoleNamePositions) {
          this.positions = data.shortRoleNamePositions
            .map(x => {
              return { label: this.translate.instant(`common.positions.${x}`), value: x };
            });
        }
        this.spinner.hideSpinner();
      });
    });
  }

  public ngOnDestroy(): void {
    this.unsubscriber.next();
    this.unsubscriber.complete();
  }

  public loadNextPage(): void {
    this.loading = true;
    this.data.query<ShortRoleNameListQuery>(Queries.shortRoleNameQuery, {
        serverId: this.server.id, filter: this.filter, page: this.page, sort: this.sort
      },
      (data: Query) => {
        return data.servers[0];
      }
    ).subscribe(data => {
      this.totalRecords = data.shortRoleNameCount;
      this.shortRoleNames = data.shortRoleNames;
      this.spinner.hideSpinner();
      this.loading = false;
    });
  }

  public setFilterForm(): void {
    this.filterForm = this.fb.group({
      id: new FormControl<number | null>(null),
      shortName: new FormControl<string | null>(null),
      roleId: new FormControl<string | null>(null),
      position: new FormControl<string | null>(null)
    });

    this.filterForm.valueChanges.pipe(
      takeUntil(this.unsubscriber),
      debounceTime(600)
    ).subscribe(changes => {
      if (changes.id) {
        this.filter.id = changes.id;
      } else {
        this.filter.id = undefined;
      }

      if (changes.shortName) {
        this.filter.shortName = changes.shortName;
      } else {
        this.filter.shortName = undefined;
      }

      if (changes.roleId) {
        this.filter.roleId = changes.roleId;
      } else {
        this.filter.roleId = undefined;
      }

      if (changes.position) {
        this.filter.position = changes.position;
      } else {
        this.filter.position = undefined;
      }

      if (this.page.pageSize)
        this.page.pageSize = 10;

      if (this.page.pageIndex)
        this.page.pageIndex = 0;

      this.loadNextPage();
    });
  }

  public newShortRoleNameTemplate: ShortRoleName = {
    createdAt: "",
    modifiedAt: ""
  };

  public nextPage(event: LazyLoadEvent): void {
    this.page.pageSize = event.rows ?? 0;
    if (event.first != null && event.rows != null)
      this.page.pageIndex = event.first / event.rows;
    this.sort.sortColumn = event.sortField ?? undefined;
    this.sort.sortDirection = event.sortOrder === 1 ? SortDirection.ASC : event.sortOrder === -1 ? SortDirection.DESC : SortDirection.ASC;

    this.loadNextPage();
  }

  public resetFilters(): void {
    this.filterForm.reset();
  }

  public onRowEditInit(table: Table, user: User, index: number): void {
    this.clonedShortRoleNames[index] = { ...user };
  }

  public override onRowEditSave(newShortRoleName: ShortRoleName, index: number): void {
    if (this.isEditingNew && JSON.stringify(newShortRoleName) === JSON.stringify(this.newShortRoleNameTemplate)) {
      this.isEditingNew = false;
      this.shortRoleNames.splice(index, 1);
      return;
    }

    if (!newShortRoleName.id && !this.isEditingNew || !newShortRoleName.shortName && !newShortRoleName.roleId && !newShortRoleName?.position) {
      return;
    }

    if (this.isEditingNew) {
      this.spinner.showSpinner();
      this.data.mutation<ShortRoleNameMutationResult>(Mutations.createShortRoleName, {
          shortName: newShortRoleName.shortName,
          roleId: newShortRoleName.roleId,
          position: newShortRoleName.position,
          serverId: this.server.id
        }
      ).pipe(catchError(err => {
        this.isEditingNew = false;
        this.spinner.hideSpinner();
        return throwError(err);
      })).subscribe(result => {
        this.isEditingNew = false;
        this.spinner.hideSpinner();
        this.toastService.success(this.translate.instant("view.server.short_role_names.message.short_role_names_create"), this.translate.instant("view.server.short_role_names.message.short_role_names_create_d", { name: result.shortRoleName.createShortRoleName?.shortName }));
        this.loadNextPage();
      });
      return;
    }

    this.spinner.showSpinner();
    this.data.mutation<UpdateUserMutationResult>(Mutations.updateShortRoleName, {
        id: newShortRoleName.id,
        shortName: newShortRoleName.shortName,
        roleId: newShortRoleName.roleId,
        position: newShortRoleName.position
      }
    ).pipe(catchError(err => {
      this.spinner.hideSpinner();
      return throwError(err);
    })).subscribe(_ => {
      this.spinner.hideSpinner();
      this.toastService.success(this.translate.instant("view.server.short_role_names.message.short_role_names_update"), this.translate.instant("view.server.short_role_names.message.short_role_names_update_d", { name: newShortRoleName.shortName }));
      this.loadNextPage();
    });

  }

  public onRowEditCancel(index: number): void {
    if (this.isEditingNew) {
      this.shortRoleNames.splice(index, 1);
      delete this.clonedShortRoleNames[index];
      this.isEditingNew = false;
      return;
    }

    this.shortRoleNames[index] = this.clonedShortRoleNames[index];
    delete this.clonedShortRoleNames[index];
  }

  public deleteShortRoleName(shortRoleName: ShortRoleName): void {
    this.confirmDialog.confirmDialog(
      this.translate.instant("view.server.short_role_names.message.short_role_names_delete"), this.translate.instant("view.server.short_role_names.message.short_role_names_delete_q", { name: shortRoleName.shortName }),
      () => {
        this.spinner.showSpinner();
        this.data.mutation<ShortRoleNameMutationResult>(Mutations.deleteShortRoleName, {
            id: shortRoleName.id
          }
        ).pipe(catchError(err => {
          this.spinner.hideSpinner();
          return throwError(err);
        })).subscribe(l => {
          this.spinner.hideSpinner();
          this.toastService.success(this.translate.instant("view.server.short_role_names.message.short_role_names_deleted"), this.translate.instant("view.server.short_role_names.message.short_role_names_deleted_d", { name: shortRoleName.shortName }));
          this.loadNextPage();
        });
      });
  }

  public addShortRoleName(table: Table): void {
    const newShortRoleName = JSON.parse(JSON.stringify(this.newShortRoleNameTemplate));

    this.shortRoleNames = [newShortRoleName, ...this.shortRoleNames];

    table.initRowEdit(newShortRoleName);

    const index = this.shortRoleNames.findIndex(l => l.id == newShortRoleName.id);
    this.onRowEditInit(table, newShortRoleName, index);

    this.isEditingNew = true;
  }

}
